/* **************************************************************************

  RGBTOOLS.C, a library to fascilitate plotting on Gdk rgb data structures.
  
Copyright (C) 2000 John Conner Atkeson

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public
License as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with this library; if not, write to the Free
Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

For further information, contact:
John Atkeson jcatkeson@juno.com
210 Drummond Place, Apt 19, Norfolk Va. 23507 USA

************************************************************************** */

#include "rgbtools.h"
/*
#define IMAGE_WIDTH  256
#define IMAGE_HEIGHT 256
*/
#if 1
  #define rgb_hline(buf,x,y,width,gs) memcpy(&(buf[((y * wdt)+x) * pixelsize]),gs->rascolor,width*pixelsize);
  #define rgb_vline(buf,x,y,height,gs) \
  {\
    int ycnt,starty,endy; \
    starty = (y*wdt+x)*pixelsize; \
    endy = starty+(height*bytewdt); \
    for (ycnt=starty;ycnt<endy;ycnt+=bytewdt){ \
      memcpy(&(buf[ycnt]),gs->rascolor,pixelsize); \
    } \
  }
  
/*    printf("rgb_hline line number:%i",__LINE__); */
/*    printf("rgb_vline line number:%i",__LINE__); */
#else
void rgb_hline(guchar *buf,int wdt,int x,int y,int width,gsettingsT *gs)
{
  memcpy(&(buf[((y * wdt)+x) * pixelsize]),gs->rascolor,width*pixelsize);
}

void rgb_vline(guchar *buf,int wdt,int x,int y,int width,gsettingsT *gs)
{
  int scrwdt3=wdt * pixelsize;
  int ycnt,endy,xnew,ynew;
  xnew = x*pixelsize;
  ynew = y*scrwdt3+xnew;
  endy = ynew+(height*scrwdt3);
  if (ycnt>){
  }
  for (ycnt=ynew;ycnt<endy;ycnt+=scrwdt3){
    memcpy(&(buf[ycnt]),gs->rascolor,pixelsize);
  }
}
  
/*    printf("rgb_hline line number:%i",__LINE__); */
/*    printf("rgb_vline line number:%i",__LINE__); */
#endif

typedef struct boxT{
  double rx1,ry1,rx2,ry2,Wdt,Hgt;
}boxT;

/* enum ClipState {Outside,Clipped,Inside}; moved to .h */

rgbpadT pad={0,0,0,0,NULL};

void rgb_setpad(rgbpadT *pad,int wdt,int hgt)
/* ***********************************************************
*********************************************************** */
{
  int bytewdt,bytesize;
  if (pad->ras) g_free(pad->ras);
  bytewdt=wdt * pixelsize;/* same as rowstride. */
  bytesize=bytewdt * hgt;
  if (bytesize<=0){
    bytesize=0; bytewdt=0; hgt=0; wdt=0;
  }
  pad->ras=g_malloc(bytesize);
  pad->bytewdt=bytewdt; pad->bytesize=bytesize;
  pad->wdt=wdt; pad->hgt=hgt;
}/* rgb_setpad */
void rgb_freepad(rgbpadT *pad)
/* ***********************************************************
*********************************************************** */
{
  g_free(pad->ras);
  pad->bytesize=0;
  pad->ras=NULL;
  pad->wdt=0; pad->hgt=0;
}/* rgb_freepad */

void rgb_fillzone(rgbpadT *pad,int x1,int y1,int x2, int y2,int r,int g,int b)
/* ***********************************************************
*  Fill a region with a given RGB color.
*********************************************************** */
{
  int xcnt,xend,ycnt,ystart,yend, byteboxwdt;
  int bytescrwdt=pad->wdt * pixelsize;
  guchar *ras=pad->ras;

  /* clipping */
  if (x1<0) x1=0; if (x2>pad->wdt) x2=pad->wdt; 
  if (y1<0) y1=0; if (y2>pad->hgt) y2=pad->hgt;
  byteboxwdt = (x2-x1)*pixelsize;

  x1 *= pixelsize;
  ystart = y1*bytescrwdt+x1; yend = y2*bytescrwdt+x1;

  for (ycnt=ystart;ycnt<yend;ycnt+=bytescrwdt){ 
    xend=ycnt+byteboxwdt; xcnt=ycnt;
    while (xcnt<xend){ /* set each pixel to RGB value */
      ras[xcnt]=r; xcnt++; ras[xcnt]=g; xcnt++; ras[xcnt]=b; xcnt++;
//      *ras++ = r; *ras++ = g; *ras++ = b;
    }
  }
}/* rgb_fillzone */

void rgb_gelbox(rgbpadT *pad,int x1,int y1,int x2, int y2,int r,int g,int b)
/* ***********************************************************
*  Average a region with a given RGB color.
*********************************************************** */
{
  int xcnt,xend,ycnt,ystart,yend, byteboxwdt;
  int bytescrwdt=pad->wdt * pixelsize;
  guchar *ras=pad->ras;

  /* clipping */
  if (x1<0) x1=0; if (x2>pad->wdt) x2=pad->wdt; 
  if (y1<0) y1=0; if (y2>pad->hgt) y2=pad->hgt;
  byteboxwdt = (x2-x1)*pixelsize;

  x1 *= pixelsize;
  ystart = y1*bytescrwdt+x1; yend = y2*bytescrwdt+x1;

  for (ycnt=ystart;ycnt<yend;ycnt+=bytescrwdt){ 
    xend=ycnt+byteboxwdt; xcnt=ycnt;
    while (xcnt<xend){ /* average each pixel with RGB value */
      ras[xcnt]=(ras[xcnt]+r)>>1; xcnt++;
      ras[xcnt]=(ras[xcnt]+g)>>1; xcnt++;
      ras[xcnt]=(ras[xcnt]+b)>>1; xcnt++;
    }
  }
}/* rgb_gelbox */

void rgb_gelcircle(rgbpadT *pad,int xcen,int ycen,int radius, int r,int g,int b)
/* ***********************************************************
*  Average a circular region with a given RGB color.
*********************************************************** */
{
  int xcnt,xstart,xend, ycnt,ystart,yend,rely;
  int bytescrwdt=pad->wdt * pixelsize;
  int halfwdt,padwdt=pad->wdt;
  guchar *ras=pad->ras;
  int rsq=radius*radius;

  ystart=ycen-radius; yend=ycen+radius;

  /* Y-clipping */
  if (ystart<0) ystart=0; if (yend>pad->hgt) yend=pad->hgt;

  for (ycnt=ystart;ycnt<yend;ycnt++){
    rely=ycnt-ycen;
    halfwdt = sqrt(rsq-(rely*rely));
    xstart=xcen-halfwdt; xend=xcen+halfwdt;

    /* X-clipping */
    if (xstart<0) xstart=0; if (xend>padwdt) xend=padwdt;

    /* now adjust for real array index. */
    xend=(xend-xstart)*pixelsize;/* x-length, actually. */
    xstart = (ycnt*bytescrwdt) + (xstart*pixelsize); 
    xend+=xstart; xcnt=xstart;
    while (xcnt<xend){ /* average each pixel with RGB value */
      ras[xcnt]=(ras[xcnt]+r)>>1; xcnt++;
      ras[xcnt]=(ras[xcnt]+g)>>1; xcnt++;
      ras[xcnt]=(ras[xcnt]+b)>>1; xcnt++;
    }
  }
}/* rgb_gelcircle */

void rgb_gelcircle_more(rgbpadT *pad,int xcen,int ycen,int radius, int r,int g,int b)
/* ***********************************************************
*   Like gelcircle, but the RGB color is 3/4 of the resulting
* pixel color.
*********************************************************** */
{
  int xcnt,xstart,xend, ycnt,ystart,yend,rely;
  int bytescrwdt=pad->wdt * pixelsize;
  int halfwdt,padwdt=pad->wdt;
  guchar *ras=pad->ras;
  int rsq=radius*radius;

  ystart=ycen-radius; yend=ycen+radius;
  r*=3; g*=3; b*=3;

  /* Y-clipping */
  if (ystart<0) ystart=0; if (yend>pad->hgt) yend=pad->hgt;

  for (ycnt=ystart;ycnt<yend;ycnt++){
    rely=ycnt-ycen;
    halfwdt = sqrt(rsq-(rely*rely));
    xstart=xcen-halfwdt; xend=xcen+halfwdt;

    /* X-clipping */
    if (xstart<0) xstart=0; if (xend>padwdt) xend=padwdt;

    /* now adjust for real array index. */
    xend=(xend-xstart)*pixelsize;/* x-length, actually. */
    xstart = (ycnt*bytescrwdt) + (xstart*pixelsize); 
    xend+=xstart; xcnt=xstart;
    while (xcnt<xend){ /* average each pixel with RGB value */
      ras[xcnt]=(ras[xcnt]+r)>>2; xcnt++;
      ras[xcnt]=(ras[xcnt]+g)>>2; xcnt++;
      ras[xcnt]=(ras[xcnt]+b)>>2; xcnt++;
    }
  }
}/* rgb_gelcircle_more */


void setbox(boxT *Bx,double rx1,double ry1,double rx2,double ry2)
/* **************************************************************
************************************************************** */
{
  Bx->rx1=rx1; Bx->ry1=ry1;
  Bx->rx2=rx2; Bx->ry2=ry2;
  Bx->Wdt=rx2-rx1; Bx->Hgt=ry2-ry1;
}

void rgb_setcolor(gsettingsT *gs,guchar red,guchar green,guchar blue,int linewidth)
/* **************************************************************
*  fills gsettings raster line with the appropriate color.
************************************************************** */
{
  rgbT rgb;
  int cnt;

  rgb[0]=red; rgb[1]=green; rgb[2]=blue;
  memcpy(gs->example,rgb,pixelsize);
/*  if (gs->rascolor!=NULL){ g_free(gs->rascolor); } */
  gs->rascolor=g_new(guchar,linewidth*pixelsize);

  linewidth*=pixelsize;
  for (cnt=0;cnt<linewidth;cnt+=pixelsize){
    memcpy(&(gs->rascolor[cnt]),rgb,pixelsize);
  }
  gs->raswidth=linewidth;
}/* rgb_setcolor */
void rgb_freecolor(gsettingsT *gs)
/* **************************************************************
*  Frees memory allocated for gsettings raster line,
* but leaves example color intact.
************************************************************** */
{
  g_free(gs->rascolor);
  gs->rascolor=NULL; 
  gs->raswidth=0;
}/* rgb_freecolor */

void rgb_adapt_gsettings(gsettingsT *gs,int newlinewidth)
/* **************************************************************
*  Adapt painting raster line to new width.
************************************************************** */
{
  int cnt,oldsize;
  rgbT rgb;
  
  memcpy(rgb,gs->example,pixelsize);

  oldsize=gs->raswidth*pixelsize;
  gs->raswidth=newlinewidth;
  newlinewidth*=pixelsize;
  if (gs->rascolor!=NULL){
    gs->rascolor=g_realloc(gs->rascolor,newlinewidth);
  }else{
    gs->rascolor=g_new(guchar,newlinewidth);
  }
  for (cnt=oldsize;cnt<newlinewidth;cnt+=pixelsize){
    memcpy(&(gs->rascolor[cnt]),rgb,pixelsize);
  }
}/* rgb_adapt_gsettings */

void rgb_create_palette(paletteT *palette,int numcolors)
/* **************************************************************
************************************************************** */
{
  int cnt;
  palette->numcolors=numcolors;
  palette->colors=calloc(sizeof(gsettingsT),numcolors);
  for (cnt=0;cnt<numcolors;cnt++){
    memset(&(palette->colors[cnt]),0,sizeof(gsettingsT));
  }
}/* rgb_create_palette */
void rgb_free_palette(paletteT *palette)
/* **************************************************************
************************************************************** */
{
  int cnt;
  for (cnt=0;cnt<palette->numcolors;cnt++){
    rgb_freecolor(&(palette->colors[cnt]));
  }
  free(palette->colors);
  palette->numcolors=0;
}/* rgb_free_palette */
void rgb_pack_palette(paletteT *palette)
/* **************************************************************
*  Memory housekeeping; reduce all the colors' paintlines
* to a mere 1 pixel.
************************************************************** */
{
  int cnt;
  int numcolors=palette->numcolors;
  for (cnt=0;cnt<numcolors;cnt++){
    rgb_adapt_gsettings(&(palette->colors[cnt]),1);
  }
}/* rgb_pack_palette */
void rgb_palette_set_spectral(paletteT *palette,int numcolors)
/* **************************************************************
************************************************************** */
{
  guchar r=0,g=0,b=0;
  int cnt,halfnum=numcolors>>1;

  rgb_create_palette(palette,numcolors);

  /* first half fade from red to green. */
  for (cnt=0;cnt<halfnum;cnt++){
    g=(cnt*255)/halfnum;  r=255-g;
    rgb_setcolor(&(palette->colors[cnt]),r,g,b,1);
  }
  /* second half fade from green to blue. */
  for (cnt=halfnum;cnt<numcolors;cnt++){
    b=((cnt-halfnum)*255)/halfnum;  g=255-b;
    rgb_setcolor(&(palette->colors[cnt]),r,g,b,1);
  }
}/* rgb_palette_set_spectral */


void rgb_putpixel(rgbpadT *rgbpad,int x,int y,guchar *rgb)
/* *************************************************************
*  One-pixel paint.
************************************************************* */
{
  if (0<=x && x<rgbpad->wdt){
    if (0<=y && y<rgbpad->hgt){
      memcpy(&(rgbpad->ras[((y * rgbpad->wdt)+x) * pixelsize]),rgb,pixelsize);
    }
  }
}/* rgb_putpixel */

void clipline(
  int *lx1,int *ly1,int *lx2,int *ly2,
  int rx1,int ry1,int rx2,int ry2,
  int *ClipState)
/* **************************************************************
*  Clips lines to edge of box.  Assumes box is sorted.
************************************************************** */
{
  double Slope,Base;
  bool LineClipped=false;

  double Lx1=*lx1,Ly1=*ly1,Lx2=*lx2,Ly2=*ly2;

  *ClipState = Outside;

  /*  X - Constrain.  */
  if (!( ((Lx1<rx1) && (Lx2<rx1)) ||
          ((Lx1>rx2) && (Lx2>rx2)) )){

    /*  Calculate slope, base for Y==MX+B.  */
    if (Lx1==Lx2 ){
      Slope = maxlong; /* Infinity. */
    }else{
      Slope = (Ly2-Ly1)/(Lx2-Lx1);
    }
    Base = Ly1-(Lx1*Slope);

    if ((Lx1<rx1) ){  /*  Take intersection with rx1. */
      Lx1 = rx1;
      Ly1 = (Slope*Lx1)+Base;
      LineClipped = true;
    }else{
      if ((Lx1>rx2) ){  /*  Take intersection with rx2. */
        Lx1 = rx2;
        Ly1 = (Slope*Lx1)+Base;
        LineClipped = true;
      }
    }

    /*  Second endpoint X.  */
    if ((Lx2<rx1) ){  /*  Take intersection with rx1. */
      Lx2 = rx1;
      Ly2 = (Slope*Lx2)+Base;
      LineClipped = true;
    }else{
      if ((Lx2>rx2) ){  /*  Take intersection with rx2. */
        Lx2 = rx2;
        Ly2 = (Slope*Lx2)+Base;
        LineClipped = true;
      }
    }

    /*  Y - Constrain.  */
    if (!( ((Ly1<ry1) && (Ly2<ry1)) ||
            ((Ly1>ry2) && (Ly2>ry2)) )){
      /*  If Ycross out of range, or Ly1 out of range anyway.  */
      if ((Ly1<ry1) ){ /*  Take intersection with ry1. */
        Ly1 = ry1;
        Lx1 = (ry1-Base)/Slope;
        LineClipped = true;
      }else{
        if ((Ly1>ry2) ){ /*  Take intersection with ry2. */
          Ly1 = ry2;
          Lx1 = (ry2-Base)/Slope;
          LineClipped = true;
        }
      }

      /*  Second endpoint Y.  */
      /*  If Ycross out of range, or Ly2 out of range anyway.  */
      if ((Ly2<ry1) ){ /*  Take intersection with ry1. */
        Ly2 = ry1;
        Lx2 = (ry1-Base)/Slope;
        LineClipped = true;
      }else{
        if ((Ly2>ry2) ){ /*  Take intersection with ry2. */
          Ly2 = ry2;
          Lx2 = (ry2-Base)/Slope;
          LineClipped = true;
        }
      }
      if (LineClipped ){
        *ClipState = Clipped;
      }else{
        *ClipState = Inside;
      }
    } /*  End if inside Y-range.  */
  } /*  End if inside X-range.  */

  *lx1=Lx1; *ly1=Ly1; *lx2=Lx2; *ly2=Ly2;
}  /*  End of clipline.  */

void internal_Clipline(int *lx1,int *ly1,int *lx2,int *ly2,int wdt,int hgt,int *ClipState)
/* **************************************************************
*  Clips lines to edge of box, left=0, top=0, right=wdt, bottom=hgt. 
************************************************************** */
{
  double Slope,Base;
  bool LineClipped=false;

  double Lx1=*lx1,Ly1=*ly1,Lx2=*lx2,Ly2=*ly2;

  *ClipState = Outside;

  /*  X - Constrain.  */
  if (!( ((Lx1<0) && (Lx2<0)) || ((Lx1>wdt) && (Lx2>wdt)) )){

    /*  Calculate slope, base for Y==MX+B.  */
    if (Lx1==Lx2 ){
      Slope = maxlong; /* Infinity. */
    }else{
      Slope = (Ly2-Ly1)/(Lx2-Lx1);
    }
    Base = Ly1-(Lx1*Slope);

    if ((Lx1<0) ){  /*  Take intersection with rx1. */
      Lx1 = 0; Ly1 = Base;
      LineClipped = true;
    }else{
      if ((Lx1>wdt) ){  /*  Take intersection with rx2. */
        Lx1 = wdt;
        Ly1 = (Slope*Lx1)+Base;
        LineClipped = true;
      }
    }

    /*  Second endpoint X.  */
    if ((Lx2<0) ){  /*  Take intersection with rx1. */
      Lx2 = 0; Ly2 = Base;
      LineClipped = true;
    }else{
      if ((Lx2>wdt) ){  /*  Take intersection with rx2. */
        Lx2 = wdt;
        Ly2 = (Slope*Lx2)+Base;
        LineClipped = true;
      }
    }

    /*  Y - Constrain.  */
    if (!( ((Ly1<0) && (Ly2<0)) || ((Ly1>hgt) && (Ly2>hgt)) )){
      /*  If Ycross out of range, or Ly1 out of range anyway.  */
      if ((Ly1<0) ){ /*  Take intersection with ry1. */
        Ly1 = 0;
        Lx1 = -Base/Slope;
        LineClipped = true;
      }else{
        if ((Ly1>hgt) ){ /*  Take intersection with ry2. */
          Ly1 = hgt;
          Lx1 = (hgt-Base)/Slope;
          LineClipped = true;
        }
      }

      /*  Second endpoint Y.  */
      /*  If Ycross out of range, or Ly2 out of range anyway.  */
      if ((Ly2<0) ){ /*  Take intersection with box y-min. */
        Ly2 = 0;
        Lx2 = -Base/Slope;
        LineClipped = true;
      }else{
        if (Ly2>hgt){ /*  Take intersection with box y-max. */
          Ly2 = hgt;
          Lx2 = (hgt-Base)/Slope;
          LineClipped = true;
        }
      }
      if (LineClipped){
        *ClipState = Clipped;
      }else{
        *ClipState = Inside;
      }
    } /*  End if inside Y-range.  */
  } /*  End if inside X-range.  */

  *lx1=Lx1; *ly1=Ly1; *lx2=Lx2; *ly2=Ly2;
}  /*  End of internal_Clipline.  */

void rgb_drawline(rgbpadT *pad,gsettingsT *gs,int x1,int y1,int x2,int y2)
/* ***********************************************************
*  Quick line drawing, rougly based on Bresenham's line
* drawing algorithm.
*********************************************************** */
{
  long aslope,xideal,yideal;
  int xinc0,xinc1,yinc0,yinc1;/* major axis increments. */
  int extra,dx,dy,x,y,temp;
  int ClipState=Outside;
  int bytewdt=pad->bytewdt;
  int wdt=pad->wdt,hgt=pad->hgt;
  guchar *buf=pad->ras;

  dx=abs(x2-x1); dy=abs(y2-y1);
  if (dx==0 && dy==0){/* one pixel */
    if (0<=x1 && x1<wdt){
      if (0<=y1 && y1<hgt){
        rgb_hline(buf,x1,y1,1,gs);
      }
    }
    return;
  }

  internal_Clipline(&x1,&y1,&x2,&y2,wdt-1,hgt-1,&ClipState);
  if (ClipState==Outside){
    return;
  }
  dx=abs(x2-x1); dy=abs(y2-y1);
#if 0
  if (dx==0 && dy==0){/* one pixel */
    rgb_hline(buf,x1,y1,1,gs); return;
  }
#endif
  if (dx>dy){
/* ***********************************************************
*  Mostly horizontal, dx>dy
*********************************************************** */
    if (gs->raswidth<wdt){
      rgb_adapt_gsettings(gs,wdt);
    }

    if (x1>x2){/* Sort ends by x. */
      temp=x1; x1=x2; x2=temp; temp=y1; y1=y2; y2=temp;
    }

    if (dy==0){
      rgb_hline(buf,x1,y1,dx,gs); return;
    }

    aslope=(dx<<8)/dy; xinc0=aslope>>8; xinc1=xinc0+1;
    xideal=x1<<8; x=x1; y=y1;

    if (y2>y1){/* line is ramping up. */

      while (y<y2){
        if ( (x<<8) > xideal){
          rgb_hline(buf,x,y,xinc0,gs);
          x+=xinc0;
        }else{
          rgb_hline(buf,x,y,xinc1,gs);
          x+=xinc1;
        }
        xideal+=aslope; y++;
      }
      extra=x2-x;
      if (extra>0){
        rgb_hline(buf,x,y,extra,gs);
      }
    }else if (y2<y1){/* line is ramping down */

      while (y>y2){
        if ( (x<<8) > xideal){
          rgb_hline(buf,x,y,xinc0,gs);
          x+=xinc0;
        }else{
          rgb_hline(buf,x,y,xinc1,gs);
          x+=xinc1;
        }
        xideal+=aslope; y--;
      }
      extra=1+x2-x;
      if (extra>0){
        rgb_hline(buf,x,y,extra,gs);
      }
    }else{/* line is true horizontal */
      rgb_hline(buf,x1,y1,x2-x1,gs);
    }
  }else{
/* ***********************************************************
*  Mostly vertical, dy>=dx
*********************************************************** */
    if (y1>y2){/* Sort ends by y. */
      temp=x1; x1=x2; x2=temp; temp=y1; y1=y2; y2=temp;
    }

    if (dx==0){
      rgb_vline(buf,x1,y1,dy,gs);
      return;
    }

    aslope=(dy<<8)/dx; yinc0=aslope>>8; yinc1=yinc0+1;
    yideal=y1<<8; x=x1; y=y1;

    if (x2>x1){/* line is ramping top-rightward. */

      while (x<x2){
        if ( (y<<8) > yideal){
          rgb_vline(buf,x,y,yinc0,gs);
          y+=yinc0;
        }else{
          rgb_vline(buf,x,y,yinc1,gs);
          y+=yinc1;
        }
        yideal+=aslope; x++;
      }
      extra=y2-y;
      if (extra>0){
        rgb_vline(buf,x,y,extra,gs);
      }
    }else if (x2<x1){/* line is ramping top-leftward. */

      while (x>x2){
        if ( (y<<8) > yideal){
          rgb_vline(buf,x,y,yinc0,gs);
          y+=yinc0;
        }else{
          rgb_vline(buf,x,y,yinc1,gs);
          y+=yinc1;
        }
        yideal+=aslope; x--;
      }
      extra=y2-y;
      if (extra>0){
        rgb_vline(buf,x,y,extra,gs);
      }

    }else{/* line is true vertical */
      rgb_vline(buf,x1,y1,y2-y1,gs);
    }

  }
}/* rgb_drawline */


#define prec 8  /* fractional precision of integer. */
void rgb_drawline2(rgbpadT *pad,gsettingsT *gs,int lx1,int ly1,int lx2,int ly2)
/* ***********************************************************
*  Quick line drawing, rougly based on Bresenham's line
* drawing algorithm, with clipping added.
*********************************************************** */
{
  long aslope,xideal,yideal;
  int xinc0,xinc1,yinc0,yinc1;/* major axis increments. */
  int extra,dx,dy,x,y,temp;
  guchar *buf=pad->ras;
  int ClipState=Outside;
  int bytewdt=pad->bytewdt;
  int wdt=pad->wdt-1;
  int hgt=pad->hgt-1;

  long Slope,Base;
  bool LineClipped=false;

  dx=abs(lx2-lx1); dy=abs(ly2-ly1);
  if (dx==0 && dy==0){/* one pixel */
    if (0<=lx1 && lx1<=wdt){
      if (0<=ly1 && ly1<=hgt){
        rgb_hline(buf,lx1,ly1,1,gs);
      }
    }
    return;
  }

#if 1

  /*  X - Constrain.  */
  if (!( ((lx1<0) && (lx2<0)) || ((lx1>wdt) && (lx2>wdt)) )){

    /*  Calculate slope, base for Y=MX+B.  */
    if (lx1==lx2 ){
      Slope = maxlong; /* Infinity. */
    }else{
      Slope = (dy<<prec)/dx;
    }
    Base = ly1-((lx1*Slope)>>prec);

    if ((lx1<0) ){  /*  Take intersection with rx1. */
      lx1 = 0; ly1 = Base;
      LineClipped = true;
    }else{
      if ((lx1>wdt) ){  /*  Take intersection with rx2. */
        lx1 = wdt;
        ly1 = ((Slope*lx1)>>prec)+Base;
        LineClipped = true;
      }
    }

    /*  Second endpoint X.  */
    if ((lx2<0) ){  /*  Take intersection with rx1. */
      lx2 = 0; ly2 = Base;
      LineClipped = true;
    }else{
      if ((lx2>wdt) ){  /*  Take intersection with rx2. */
        lx2 = wdt;
        ly2 = ((Slope*lx2)>>prec)+Base;
        LineClipped = true;
      }
    }

    /*  Y - Constrain.  */
    if (!( ((ly1<0) && (ly2<0)) || ((ly1>hgt) && (ly2>hgt)) )){
      /*  If Ycross out of range, or ly1 out of range anyway.  */
      if ((ly1<0) ){ /*  Take intersection with ry1. */
        ly1 = 0;
        lx1 = -( (Base<<prec)/Slope)>>prec;
        LineClipped = true;
      }else{
        if ((ly1>hgt) ){ /*  Take intersection with ry2. */
          ly1 = hgt;
          lx1 = (((hgt-Base)<<prec)/Slope) >> prec;
          LineClipped = true;
        }
      }

      /*  Second endpoint Y.  */
      /*  If Ycross out of range, or ly2 out of range anyway.  */
      if ((ly2<0) ){ /*  Take intersection with box y-min. */
        ly2 = 0;
        lx2 = -( (Base<<prec)/Slope)>>prec;
        LineClipped = true;
      }else{
        if (ly2>hgt){ /*  Take intersection with box y-max. */
          ly2 = hgt;
          lx2 = (((hgt-Base)<<prec)/Slope) >> prec;
          LineClipped = true;
        }
      }
      if (LineClipped){
        ClipState = Clipped;
      }else{
        ClipState = Inside;
      }
    } /*  End if inside Y-range.  */
  } /*  End if inside X-range.  */

#endif

  if (ClipState==Outside){
    return;
  }

  dx=abs(lx2-lx1); dy=abs(ly2-ly1);
  if (dx==0 && dy==0){/* one pixel */
    rgb_hline(buf,lx1,ly1,1,gs); return;
  }

  if (dx>dy){
/* ***********************************************************
*  Mostly horizontal, dx>dy
*********************************************************** */
    if (gs->raswidth<wdt){
      rgb_adapt_gsettings(gs,wdt);
    }

    if (lx1>lx2){/* Sort ends by x. */
      temp=lx1; lx1=lx2; lx2=temp; temp=ly1; ly1=ly2; ly2=temp;
    }

    if (dy==0){
      rgb_hline(buf,lx1,ly1,dx,gs); return;
    }

    aslope=(dx<<prec)/dy;
    xinc0=aslope>>prec; xinc1=xinc0+1;
    xideal=lx1<<prec; x=lx1; y=ly1;

    if (ly2>ly1){/* line is ramping up. */

      while (y<ly2){
        if ( (x<<prec) > xideal){
          rgb_hline(buf,x,y,xinc0,gs);
          x+=xinc0;
        }else{
          rgb_hline(buf,x,y,xinc1,gs);
          x+=xinc1;
        }
        xideal+=aslope; y++;
      }
      extra=lx2-x;
      if (extra>0){
        rgb_hline(buf,x,y,extra,gs);
      }
    }else if (ly2<ly1){/* line is ramping down */

      while (y>ly2){
        if ( (x<<prec) > xideal){
          rgb_hline(buf,x,y,xinc0,gs);
          x+=xinc0;
        }else{
          rgb_hline(buf,x,y,xinc1,gs);
          x+=xinc1;
        }
        xideal+=aslope; y--;
      }
      extra=1+lx2-x;
      if (extra>0){
        rgb_hline(buf,x,y,extra,gs);
      }
    }else{/* line is true horizontal */
      rgb_hline(buf,lx1,ly1,dx,gs);
    }
  }else{
/* ***********************************************************
*  Mostly vertical, dy>=dx
*********************************************************** */
    if (ly1>ly2){/* Sort ends by y. */
      temp=lx1; lx1=lx2; lx2=temp; temp=ly1; ly1=ly2; ly2=temp;
    }

    if (dx==0){
      rgb_vline(buf,lx1,ly1,dy,gs);
      return;
    }

    aslope=Slope;
    yinc0=aslope>>prec; yinc1=yinc0+1;
    yideal=ly1<<prec; x=lx1; y=ly1;

    if (lx2>lx1){/* line is ramping top-rightward. */

      while (x<lx2){
        if ( (y<<prec) > yideal){
          rgb_vline(buf,x,y,yinc0,gs);
          y+=yinc0;
        }else{
          rgb_vline(buf,x,y,yinc1,gs);
          y+=yinc1;
        }
        yideal+=aslope; x++;
      }
      extra=ly2-y;
      if (extra>0){
        rgb_vline(buf,x,y,extra,gs);
      }
    }else if (lx2<lx1){/* line is ramping top-leftward. */

      while (x>lx2){
        if ( (y<<prec) > yideal){
          rgb_vline(buf,x,y,yinc0,gs);
          y+=yinc0;
        }else{
          rgb_vline(buf,x,y,yinc1,gs);
          y+=yinc1;
        }
        yideal+=aslope; x--;
      }
      extra=ly2-y;
      if (extra>0){
        rgb_vline(buf,x,y,extra,gs);
      }

    }else{/* line is true vertical */
      rgb_vline(buf,lx1,ly1,ly2-ly1,gs);
    }

  }
}/* rgb_drawline2 */


void rgb_fillpad(rgbpadT *pad,guchar r,guchar g,guchar b)
/* ***********************************************************
* fill screen with one color.
*********************************************************** */
{
  intdex bytesize=pad->bytesize,wdt=pad->wdt,cnt;
  guchar *buf=pad->ras;
  gsettingsT gs={0,NULL,0,0,0};

  rgb_setcolor(&gs, r,g,b, pad->wdt);
  for (cnt=0;cnt<bytesize;cnt+=wdt){
    memcpy(&(buf[cnt]), gs.rascolor, wdt);
  }
  rgb_freecolor(&gs);
}/* rgb_fillpad */

void rgb_updatezone(GtkWidget *widget,rgbpadT *pad,int x1,int y1,int x2,int y2)
/* ***********************************************************
*   Copy a rectangular zone of an rgb to the 
* corresponding zone of a drawable.
*********************************************************** */
{
  int offset,wdt,hgt;

  /* clipping */
  if (x1<0) x1=0; if (x2>pad->wdt) x2=pad->wdt;   
  if (y1<0) y1=0; if (y2>pad->hgt) y2=pad->hgt;
  wdt=x2-x1; hgt=y2-y1;
  
  if (wdt<=0 || hgt<=0) return;/* no box left! */

  offset=((y1*pad->wdt)+x1)*pixelsize;
#if 0 /* for debugging */
  printf("window:%X, xyxy:%i, %i, %i, %i  ras[]:%i, %i, offset:%i bz:%li\n",
      widget->window, x1, y1, x2, y2, pad->ras[offset], pad->wdt * pixelsize,offset,pad->bytesize);
#endif
//  rgb_gelbox(pad,0,0,pad->wdt-1,pad->hgt-1, 75,200,75);

  gdk_draw_rgb_image (widget->window, widget->style->fg_gc[GTK_STATE_NORMAL],
                      x1, y1, wdt, hgt,
                      GDK_RGB_DITHER_MAX, &(pad->ras[offset]), pad->bytewdt);
}/* rgb_updatezone */

void rgb_clonezone2(rgbpadT *pad1,rgbpadT *pad2,int xdest,int ydest, int x1,int y1,int x2,int y2)
/* ***********************************************************
*   Copy a rectangular zone from pad2 to pad1.
*********************************************************** */
{
  intdex cnt1,cnt2,offset1,offset2,rowstride1,rowstride2,zonewdt;
  intdex endy;

  /* clipping */
  if (x1<0) x1=0; 
  if (x2>pad1->wdt) x2=pad1->wdt; if (x2>pad2->wdt) x2=pad2->wdt;
  if (y1<0) y1=0; 
  if (y2>pad1->hgt) y2=pad1->hgt; if (y2>pad2->hgt) y2=pad2->hgt;

  zonewdt=x2-x1; rowstride1=pad1->wdt; rowstride2=pad2->wdt;

  offset1=((ydest*pad1->wdt)+xdest)*pixelsize;
  offset2=((y1*pad2->wdt)+x1)*pixelsize;

  for (cnt1=offset2;cnt1<endy;cnt1+=rowstride1){
    memcpy(&(pad1->ras[cnt1]),&(pad2->ras[cnt2]),zonewdt);
    cnt2+=rowstride2;
  }
}/* rgb_clonezone2 */

void rgb_clonezone(rgbpadT *pad1,rgbpadT *pad2,int xdest,int ydest, int x1,int y1,int x2,int y2)
/* ***********************************************************
*   Copy a rectangular zone from pad2 to pad1.
*********************************************************** */
{
  int offset1,offset2,zonewdt,zonehgt;
  int ycnt1,ycnt2,wdt1,wdt2;
  int xdest2,ydest2,overhang;

  /* clipping to destination box. */
  if (xdest>pad1->wdt || ydest>pad1->hgt){/* destination beyond edge. */
    return;
  }
  if (xdest<0) { x1-=xdest; xdest=0; }
  if (ydest<0) { y1-=ydest; ydest=0; }
  zonewdt=x2-x1; zonehgt=y2-y1;

  overhang=((xdest+zonewdt)-pad1->wdt);  if (overhang>0) x2-=overhang;
  overhang=((ydest+zonehgt)-pad1->hgt);  if (overhang>0) y2-=overhang;

  /* clipping to source box. */
  if (x1<0) x1=0; if (x2>pad2->wdt) x2=pad2->wdt;
  if (y1<0) y1=0; if (y2>pad2->hgt) y2=pad2->hgt;

  zonewdt=(x2-x1)*pixelsize;
  wdt1=pad1->wdt;  wdt2=pad2->wdt;
  
  if (zonewdt>0){
    ycnt1=ydest;
    for (ycnt2=y1;ycnt2<y2;ycnt2++){
      offset2=((ycnt2*wdt2)+x1)*pixelsize;/* coming from */
      offset1=((ycnt1*wdt1)+xdest)*pixelsize;/* going to */
      memcpy(&(pad1->ras[offset1]),&(pad2->ras[offset2]),zonewdt);
      ycnt1++;
    }
  }
}/* rgb_clonezone */


void rgb_slide_pad(rgbpadT *rgbpad,int dx,int dy)
/* *************************************************************
*  Copy & move a chunk of rgbpad off toward one side.
************************************************************* */
{
  int xsrc,ysrc,xdest,ydest,patchwdt,patchhgt,maxx,maxy;
  rgbpadT newpad;
  int absdx=abs(dx),absdy=abs(dy);

  if (dx<0){ xsrc=absdx; xdest=0; }else{ xsrc=0; xdest=absdx; }
  if (dy<0){ ysrc=absdy; ydest=0; }else{ ysrc=0; ydest=absdy; }

  maxx=rgbpad->wdt; maxy=rgbpad->hgt;
  patchwdt=maxx-absdx;  patchhgt=maxy-absdy;
  rgb_setpad(&newpad,patchwdt,patchhgt);

  rgb_clonezone(&newpad,rgbpad, 0,0, xsrc,ysrc, maxx,maxy);
  rgb_clonezone(rgbpad,&newpad, xdest,ydest, 0,0,patchwdt,patchhgt);
  rgb_freepad(&newpad);
}/* rgb_slide_pad */

#if 0
void clipline(int *lx1,int *ly1,int *lx2,int *ly2, boxT *Box, int *ClipState)
/* **************************************************************
*  Clips lines to edge of box.  Assumes box is sorted.
************************************************************** */
{
  double Slope,Base;
  bool LineClipped=false;

  double Lx1=*lx1,Ly1=*ly1,Lx2=*lx2,Ly2=*ly2;

  *ClipState = Outside;

  /*  X - Constrain.  */
  if (!( ((Lx1<Box->rx1) && (Lx2<Box->rx1)) ||
          ((Lx1>Box->rx2) && (Lx2>Box->rx2)) )){

    /*  Calculate slope, base for Y==MX+B.  */
    if (Lx1==Lx2 ){
      Slope = maxlong; /* Infinity. */
    }else{
      Slope = (Ly2-Ly1)/(Lx2-Lx1);
    }
    Base = Ly1-(Lx1*Slope);

    if ((Lx1<Box->rx1) ){  /*  Take intersection with rx1. */
      Lx1 = Box->rx1;
      Ly1 = (Slope*Lx1)+Base;
      LineClipped = true;
    }else{
      if ((Lx1>Box->rx2) ){  /*  Take intersection with rx2. */
        Lx1 = Box->rx2;
        Ly1 = (Slope*Lx1)+Base;
        LineClipped = true;
      }
    }

    /*  Second endpoint X.  */
    if ((Lx2<Box->rx1) ){  /*  Take intersection with rx1. */
      Lx2 = Box->rx1;
      Ly2 = (Slope*Lx2)+Base;
      LineClipped = true;
    }else{
      if ((Lx2>Box->rx2) ){  /*  Take intersection with rx2. */
        Lx2 = Box->rx2;
        Ly2 = (Slope*Lx2)+Base;
        LineClipped = true;
      }
    }

    /*  Y - Constrain.  */
    if (!( ((Ly1<Box->ry1) && (Ly2<Box->ry1)) ||
            ((Ly1>Box->ry2) && (Ly2>Box->ry2)) )){
      /*  If Ycross out of range, or Ly1 out of range anyway.  */
      if ((Ly1<Box->ry1) ){ /*  Take intersection with ry1. */
        Ly1 = Box->ry1;
        Lx1 = (Box->ry1-Base)/Slope;
        LineClipped = true;
      }else{
        if ((Ly1>Box->ry2) ){ /*  Take intersection with ry2. */
          Ly1 = Box->ry2;
          Lx1 = (Box->ry2-Base)/Slope;
          LineClipped = true;
        }
      }

      /*  Second endpoint Y.  */
      /*  If Ycross out of range, or Ly2 out of range anyway.  */
      if ((Ly2<Box->ry1) ){ /*  Take intersection with ry1. */
        Ly2 = Box->ry1;
        Lx2 = (Box->ry1-Base)/Slope;
        LineClipped = true;
      }else{
        if ((Ly2>Box->ry2) ){ /*  Take intersection with ry2. */
          Ly2 = Box->ry2;
          Lx2 = (Box->ry2-Base)/Slope;
          LineClipped = true;
        }
      }
      if (LineClipped ){
        *ClipState = Clipped;
      }else{
        *ClipState = Inside;
      }
    } /*  End if inside Y-range.  */
  } /*  End if inside X-range.  */

  *lx1=Lx1; *ly1=Ly1; *lx2=Lx2; *ly2=Ly2;
}  /*  End of clipline.  */

#endif
