/* **************************************************************************

  TIN.C, a library to create Triangulated Irregular Networks
from xyz point data.

Copyright (C) 2000 John Conner Atkeson III

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

For further information, contact:
John Conner Atkeson III
jcatkeson at gmail dot com
http://jca3.freeshell.org
3959 Persimmon Drive #103 Fairfax Va. 22031

************************************************************************** */

#include "tinlib.h"
time_t randseed;

#define testing 0

char *vertexpnt="vertexpnt";

bool point_test(XYZType *Pt)
/* *******************************************************************
*  Just a stub to fill the user-definable filter.
* ******************************************************************* */
{
  return(true);
}

lineT *line_create(uint32 pdex1,uint32 pdex2,lstateT state)
/* *******************************************************************
* ******************************************************************* */
{
  lineT *line;
  line=malloc(sizeof(lineT));
  line->pdex1=pdex1; line->pdex2=pdex2; line->state=state;
  return(line);
}/* line_create */
void line_free(lineT *line)
/* *******************************************************************
*  Free the memory allocated to a line.
* ******************************************************************* */
{
  free(line);
}/* line_free */

void DrawLoop(brickptrT loop_wall,PListT *points,int color)
/* *******************************************************************
*  For debugging, not used.
* ******************************************************************* */
{
  XYZType *pt1,*pt2,*P;
  intdex pdex1,pdex2,brickcnt;
  brickptrT now;
  lineT *this_line;

  P=points->P;

  brickcnt=0;
  now=loop_wall;
  do{
    this_line=now->linptr;
    pdex1=this_line->pdex1; pdex2=this_line->pdex2;
    pt1=&(P[pdex1]); pt2=&(P[pdex2]);

    if (this_line->state==outer){
      color=1;
    }else{
      color=2;
    }
#if allegro
    line(screen,
      pt1->Loc[0],pt1->Loc[1], pt2->Loc[0],pt2->Loc[1], color);
#endif
    now=now->next;
    brickcnt++;
//    color++;
  }while(now!=loop_wall);
}/* DrawLoop */

void draw_triangle(PListT *points,intdex pdex1,intdex pdex2,intdex pdex3,int color)
/* *******************************************************************
* ******************************************************************* */
{
  XYZType *pt1,*pt2,*pt3;

//  color=1;
//  color%=60; if (color==0)color=1;

  pt1=&(points->P[pdex1]); pt2=&(points->P[pdex2]); pt3=&(points->P[pdex3]);
#if allegro
  line(screen,
    pt1->Loc[0],pt1->Loc[1], pt2->Loc[0],pt2->Loc[1], color);

  line(screen,
    pt3->Loc[0],pt3->Loc[1], pt2->Loc[0],pt2->Loc[1], color);

  line(screen,
    pt1->Loc[0],pt1->Loc[1], pt3->Loc[0],pt3->Loc[1], color);
#endif
}/* draw_triangle */

void tin_draw_triangles(tin_contextT *tcontext)
/* *******************************************************************
*  Draw triangles onscreen, only used for debugging with allegro.
* ******************************************************************* */
{
  triT *tri;
  flexelT *tri_tree;
  uint32 tcnt,num_triangles;

  tri_tree=&(tcontext->tri_tree);

  num_triangles=tcontext->num_triangles;
  for(tcnt=0;tcnt<num_triangles;tcnt++){
    tri=flexray_lookup(tri_tree,tcnt);
    draw_triangle(tcontext->points,tri->vtex[0],tri->vtex[1],tri->vtex[2],4);
  }
}/* tin_draw_triangles */

uint32 count_bricks(brickptrT loop_wall)
/* *******************************************************************
* ******************************************************************* */
{
  intdex brickcnt;
  brickptrT now;

  brickcnt=0; now=loop_wall;
  do{
    now=now->next; brickcnt++;
  }while(now!=loop_wall);
  return(brickcnt);
}/* count_bricks */


void save_triangle(FILE *outfile,PListT *points,intdex pdex0,intdex pdex1,intdex pdex2)
/* *******************************************************************
*  Save one triangle to a file. (Companion to tin_save_triangles.)
* ******************************************************************* */
{
  XYZType *pt0,*pt1,*pt2;
  pt0=&(points->P[pdex0]); pt1=&(points->P[pdex1]); pt2=&(points->P[pdex2]);

  fprintf(outfile,"(%1.1lf, %1.1lf, %1.1lf), ",pt0->Loc[0],pt0->Loc[1],pt0->Loc[2]);
  fprintf(outfile,"(%1.1lf, %1.1lf, %1.1lf), ",pt1->Loc[0],pt1->Loc[1],pt1->Loc[2]);
  fprintf(outfile,"(%1.1lf, %1.1lf, %1.1lf)\n",pt2->Loc[0],pt2->Loc[1],pt2->Loc[2]);
}/* save_triangle */

void tin_save_triangles(tin_contextT *tcontext,char *filename)
/* *******************************************************************
*  Save all triangles to a file.
* ******************************************************************* */
{
  triT *tri;
  flexelT *tri_tree;
  uint32 tcnt,num_triangles;
  FILE *outfile;

  tri_tree=&(tcontext->tri_tree);
  outfile=fopen(filename,"w");

  num_triangles=tcontext->num_triangles;
  for(tcnt=0;tcnt<num_triangles;tcnt++){
    tri=flexray_lookup(tri_tree,tcnt);
    save_triangle(outfile,tcontext->points,tri->vtex[0],tri->vtex[1],tri->vtex[2]);
  }
  fclose(outfile);
}/* tin_save_triangles */

void save_tinfile(tin_contextT *tcontext,char *filename)
/* *******************************************************************
*  Save whole structure of TIN to a file, including connectivity.
* ******************************************************************* */
{
  PListT *pl; XYZType *Plist,*P;
  uint32 pcnt,num_points;
  triT *tri;
  flexelT *tri_tree;
  uint32 tcnt,num_triangles;
  FILE *outfile;
  char outstr[256];

  tri_tree=&(tcontext->tri_tree);
  num_triangles=tcontext->num_triangles;

  pl=tcontext->points; num_points=pl->NumPoints; Plist=pl->P;

  outfile=fopen(filename,"w");
  fprintf(outfile,"num_points=%u num_triangles=%u\n",num_points,num_triangles);
  /* Save raw data points, in order. */
  for(pcnt=0;pcnt<num_points;pcnt++){
    P=&(Plist[pcnt]);
    sprintf(outstr,"p %1.2lf %1.2lf %1.2lf",P->Loc[0],P->Loc[1],P->Loc[2]);
    fprintf(outfile,"%s\n",outstr);
  }
  /* Save traingles as indexes of data points. */
  for(tcnt=0;tcnt<num_triangles;tcnt++){
    tri=flexray_lookup(tri_tree,tcnt);
    sprintf(outstr,"t %li %li %li",tri->vtex[0],tri->vtex[1],tri->vtex[2]);
    fprintf(outfile,"%s\n",outstr);
  }
  fclose(outfile);
}/* save_tinfile */

void store_triangle(tin_contextT *tcontext,intdex pdex0,intdex pdex1,intdex pdex2)
/* *******************************************************************
*  Store a triangle to memory in a flexray.
* ******************************************************************* */
{
  triT *tri;
  flexelT *tri_tree;
  bool success;

  tri_tree=&(tcontext->tri_tree);
  tri=malloc(sizeof(triT));
  tri->vtex[0]=pdex0; tri->vtex[1]=pdex1; tri->vtex[2]=pdex2;
  success=flexray_insert(tri_tree,tcontext->num_triangles,tri);
  if(!success){
    printf("Error, a triangle has already been stored to this index!\n");
  }
  tcontext->num_triangles++;
}/* store_triangle */

bool tin_get_triangle_full(tin_contextT *tcontext,full_triT *tri,uint32 tri_num)
/* *******************************************************************
*  Get a triangle with all its analogue coordinates.
* ******************************************************************* */
{
  triT *tridex;
  XYZType *P,*pt;
  int dimcnt,vcnt;

  tridex=flexray_lookup(&(tcontext->tri_tree),tri_num);

  if (tridex!=NULL){
    P=tcontext->points->P;
    for (vcnt=0;vcnt<3;vcnt++){
      pt=&(P[tridex->vtex[vcnt]]);
      for (dimcnt=0;dimcnt<3;dimcnt++){
        tri->vtex[vcnt][dimcnt]=pt->Loc[dimcnt];
      }
    }
    return(true);/* True if triangle found */
  }else{
    return(false);/* False if triangle not found */
  }
}/* tin_get_triangle_full */


void tin_save_outline(tin_contextT *tcontext,char *filename)
/* *******************************************************************
*   Save the loop wall to a file.
* ******************************************************************* */
{
  brickptrT loop_wall;
  PListT *points;
  XYZType *pt1,*pt2,*P;
  intdex pdex1,pdex2,brickcnt;
  brickptrT now;
  lineT *this_line;
  FILE *outfile;
  
  loop_wall=tcontext->loop_wall;
  points=tcontext->points;
  
  outfile=fopen(filename,"w");

  P=points->P;

  brickcnt=0;
  now=loop_wall;
  do{
    this_line=now->linptr;
    pdex1=this_line->pdex1; pdex2=this_line->pdex2;
    pt1=&(P[pdex1]); pt2=&(P[pdex2]);

/*    fprintf(outfile,"%1.2lf, %1.2lf\n ",pt1->Loc[0],pt1->Loc[1]); */

    fprintf(outfile,"%1.2lf %1.2lf %1.2lf - ",pt1->Loc[0],pt1->Loc[1],pt1->Loc[2]);
    fprintf(outfile,"%1.2lf %1.2lf %1.2lf\n",pt2->Loc[0],pt2->Loc[1],pt2->Loc[2]);

    now=now->next;
    brickcnt++;
  }while(now!=loop_wall);

  fclose(outfile);
}/* tin_save_outline */


/* *******************************************************************
*  Begin line list support.
* ******************************************************************* */

linelistT *linelist_create(int nlines)
/* *******************************************************************
* ******************************************************************* */
{
  int cnt;
  linelistT *lines;

  lines=malloc(sizeof(linelistT));
  lines->nlines=nlines;
  lines->L=calloc(nlines,sizeof(lineT*));
  /* probably never use this part. */
  for(cnt=0;cnt<nlines;cnt++){
//    lines->L[cnt]=malloc(sizeof(lineT));
    lines->L[cnt]=line_create(0,0,alive);
  }
  return(lines);
}
void linelist_destroy(linelistT *lines)
/* *******************************************************************
* ******************************************************************* */
{
  int cnt,nlines=lines->nlines;
  for(cnt=0;cnt<nlines;cnt++){
    free(lines->L[cnt]); lines->L[cnt]=NULL;
  }
  free(lines->L); free(lines);
}/* linelist_destroy */
void leaf_destroy(void *lines)
/* *******************************************************************
* ******************************************************************* */
{
  linelist_destroy((linelistT *)lines);
}/* leaf_destroy */

void linelist_addone(linelistT *lines,lineT *line)
/* *******************************************************************
* ******************************************************************* */
{
  int last=lines->nlines;
  lines->nlines++;
  lines->L=realloc(lines->L,lines->nlines*sizeof(lineT*));
  lines->L[last]=line;
}
lineT *linelist_remove(linelistT *lines,int linenum)
/* *******************************************************************
* ******************************************************************* */
{
  int cnt,nlines;
  lineT *retline;

//  free(lines->L[linenum]);
  retline=lines->L[linenum];
  lines->nlines--; nlines=lines->nlines;
  for (cnt=linenum;cnt<nlines;cnt++){
    lines->L[cnt]=lines->L[cnt+1];
  }
  lines->L=realloc(lines->L,nlines*sizeof(lineT*));
  return(retline);
}/* linelist_remove */
void linelist_pack(linelistT *lines,int nlines)
/* *******************************************************************
* ******************************************************************* */
{
  lines->L=realloc(lines->L,nlines*sizeof(lineT*));
  lines->nlines=nlines;
}
/* *******************************************************************
*  End line list support.
* ******************************************************************* */

void linetree_lookup(flexelT *line_tree,lineT *line,char *tag)
/* *******************************************************************
*  Remove one line from the linetree.
* ******************************************************************* */
{
  bool found=false;
  linelistT *lines;
  lineT **L;
  int linenum,nlines;

  lines=flexray_lookup(line_tree,line->pdex1);
  if (lines!=NULL){
    nlines=lines->nlines;
    L=lines->L;
  
    linenum=0;
    while(linenum<nlines){
      if(L[linenum]==line){found=true; break;}
      linenum++;
    }
  
    if(found){
      printf("I see line in %s.\n",tag);
    }else{
      printf("Dont see line in %s.\n",tag);
    }
  }else{
    printf("Dont see linelist in %s.\n",tag);
  }
}/* linetree_lookup */

void linetree_insert(flexelT *line_tree,lineT *line)
/* *******************************************************************
*  Add one line to the linetree.
* ******************************************************************* */
{
  bool found=false,success;
  linelistT *lines;
  lineT *L;
  int linenum,nlines;
  lines=flexray_lookup(line_tree,line->pdex1);
  if(lines==NULL){
    lines=linelist_create(0);
    success=flexray_insert(line_tree,line->pdex1,lines);
    if(!success){
      printf("Error in linetree_insert, linelist already there.\n");
    }
  }
  linelist_addone(lines,line);
}/* linetree_insert */

void linetree_remove(flexelT *line_tree,lineT *line)
/* *******************************************************************
*  Remove one line from the linetree.
* ******************************************************************* */
{
  bool found=false;
  linelistT *lines;
  lineT **L;
  int linenum,nlines;
  lineT *oldline;
  intdex pdex1;

  pdex1=line->pdex1;

  lines=flexray_lookup(line_tree,pdex1);
  if (lines!=NULL){
    nlines=lines->nlines;
    L=lines->L;
  
    linenum=0;
    while(linenum<nlines){
      if(L[linenum]==line){found=true; break;}
      linenum++;
    }
  
    if(found){
      oldline=linelist_remove(lines,linenum);
      if(lines->nlines<=0){
        /* all lines are gone, remove the list. */
        lines=flexray_remove(line_tree,pdex1);
        linelist_destroy(lines);
      }
      free(oldline);
    }else{
      printf("Error in linetree_remove, line not found.\n");
    }
  }else{
    printf("Error in linetree_remove, linelist not found.\n");
  }
}/* linetree_remove */

void tri_destroy(void *triangle)
/* *******************************************************************
*  Free the memory allocated to a triangle.
* ******************************************************************* */
{
  free((triT *)triangle);
//  free(triangle);
}/* tri_destroy */


uint32 tin_create_random(PListT *points,int numpoints)
/* *******************************************************************
*  Create a field of random points for testing.
* ******************************************************************* */
{
  int cnt,dimcnt;
  XYZType *pt;
  double randnum;

//  size=50;
//  randseed=955189213;
//  randseed=955234891;
  randseed=time(0);

  srand(randseed);
  for(cnt=0;cnt<8;cnt++){
    randnum=rand();
  }

  InitPList(points,numpoints);
  for(cnt=0;cnt<numpoints;cnt++){
    for(dimcnt=0;dimcnt<3;dimcnt++){
      randnum=rand()%160+20;
      randnum+=((double)(rand()%12345))/12345.6;
//      randnum=sqrt(randnum)*17.0-40.0;

      points->P[cnt].Loc[dimcnt]=randnum;
    }
    pt=&(points->P[cnt]);
#if allegro
    line(screen,pt->Loc[0],pt->Loc[1],pt->Loc[0],pt->Loc[1], 0xff);
#endif
  }
  printf("rs:%li.",randseed);
  return(randseed);
}/* tin_create_random */

bool in_salami(
  double x0,double y0,double x1,double y1,double radius,
  double xtarg,double ytarg)
/* **************************************************************
*  Returns true if xtarg,ytarg is found inside a sausage-shaped
* area.  Experimental, not currently used.
************************************************************** */
{
  bool retval;
  double dx0,dy0,dx1,dy1,dxl,dyl,rad0,rad1,linelen,dist,extension;

  dx0=(x0-xtarg); dy0=(y0-ytarg);
  rad0=sqrt(dx0*dx0 + dy0*dy0);
  if (rad0<=radius){/* Close enough to one endpoint. */
    retval=true;
//    printf("in radius0. ");
  }else{
    dx1=(x1-xtarg); dy1=(y1-ytarg);
    rad1=sqrt(dx1*dx1 + dy1*dy1);
    if (rad1<=radius){/* Close enough to the other endpoint. */
      retval=true;
//      printf("in radius1. ");
    }else{/* If all else failed, check for distance from centerline. */
      dxl=(x1-x0); dyl=(y1-y0);
      linelen=sqrt(dxl*dxl + dyl*dyl);
//      if (linelen==0.0) linelen=fudge;/* avoid divide-by-zero. */
//      dist=( ( (y0-ytarg)*dxl - (x0-xtarg)*dyl )/(linelen*linelen) )*linelen;
      dist=( (y0-ytarg)*dxl - (x0-xtarg)*dyl )/linelen;

      if (fabs(dist)<=radius){
        extension=((y0-ytarg)*(y0-y1)-(x0-xtarg)*(x1-x0))/(linelen*linelen);
        if (0.0<=extension && extension<=1.0){
          retval=true;
//          printf("near bar. ");
        }else{
          retval=false;
//          printf("outside end of bar. \n");
        }
      }else{
        retval=false;
      }
    }
  }
  return(retval);
}/* in_salami */

bool FindClosest(tin_contextT *tcontext,intdex CenterDex,intdex *NbrDex,LogicType *Filter)
/* ***************************************************************
*  Pythagorean closest to point at CenterDex.
*************************************************************** */
{
  intdex zcnt,SubjIndex,min_point_index;
  double search_radius,MinDist,Dist,border_segment_max;
  double centerX,centerY;
  VectT *paramv;/* returns whatever traversezone found. */
  CubeT szone;
  bool Found;

  paramv=tcontext->paramv;

  centerX=tcontext->points->P[CenterDex].Loc[0];
  centerY=tcontext->points->P[CenterDex].Loc[1];
  min_point_index=-1; MinDist=maxint32;/* Index of nearest neighbor. */
  search_radius=1.0;/* search radius. */
  border_segment_max=tcontext->border_segment_max;

  /* z stuff. */
  szone.L[0][2]=MinMax.L[0][2];szone.L[1][2]=MinMax.L[1][2];
  
  Found=false;
  do{
    paramv->NumPoints=0;
    /* set search bounds. */
    SetCube(centerX-search_radius,centerY-search_radius,0.0,
            centerX+search_radius,centerY+search_radius,0.0,&szone);

    TraverseZone(tcontext->sort,tcontext->points,&szone,paramv);
    for (zcnt=0;zcnt<paramv->NumPoints;zcnt++){
      SubjIndex=paramv->V[zcnt];

      {/* start test distance part. */
        /* See if this point is closer than any other found. */
        XYZType *P=&(tcontext->points->P[SubjIndex]);/* with-do for sample point. */
        Dist=Hypot(P->Loc[0]-centerX,P->Loc[1]-centerY);/* Pythagorean Distance */
        if(Dist<=search_radius){/* if within search circle. */
          if(SubjIndex!=CenterDex){/* if not myself. */
            if(Dist<MinDist){/* if closer than any other. */
              if( (*Filter)(P) ){/* user-defined filter         */
                min_point_index=SubjIndex; MinDist=Dist;
                Found=true;
              }
            }
          }
        }/* if inside radius */
      }/* end of test distance part. */

    }
    search_radius*=2.0;/* expand search zone for next. */
  }while (! (Found||(search_radius>=border_segment_max)) );
  (*NbrDex)=min_point_index;
  return(Found);
}/* End of FindClosest. */

bool FindThird(tin_contextT *tcontext,intdex pt0,intdex pt1,intdex *MinSubjIndex,LogicType *Filter)
/* ***************************************************************
* Always searches on the left hand side of line travel
* direction.
*************************************************************** */
{
  intdex zcnt,datindex,min_point_index;
  double Tx0,Ty0,Tx1,Ty1;
  double HalfDeltaX,HalfDeltaY,MidX,MidY;
  double HalfBase,Angle,RadOffset,maxarm,MaxAngle,border_segment_max;
  double search_radius,Spread,Dist;
  double centerX,centerY;
  CubeT szone;
  XYZType *P,BaseP0,BaseP1;
  bool Found;
  VectT *paramv;/* returns whatever traversezone found. */
  
  border_segment_max=tcontext->border_segment_max;

  paramv=tcontext->paramv;

  min_point_index=-1;/* Index of nearest neighbor. */
  Tx0=tcontext->points->P[pt0].Loc[0]; Ty0=tcontext->points->P[pt0].Loc[1];
  Tx1=tcontext->points->P[pt1].Loc[0]; Ty1=tcontext->points->P[pt1].Loc[1];
  
  HalfDeltaX=(Tx1-Tx0)/2.0;HalfDeltaY=(Ty1-Ty0)/2.0;
  MidX=Tx0+HalfDeltaX;MidY=Ty0+HalfDeltaY;
  HalfBase=Hypot(HalfDeltaX,HalfDeltaY);/* starter radius */
  
  min_point_index=-1;/* to scare out bugs. */
  /* set actual base points. */
  BaseP0=tcontext->points->P[pt0]; BaseP1=tcontext->points->P[pt1];
  MaxAngle=-maxint32;/* ridiculous starter comparison */

  Found=false;

  Angle=Pi/2.0;/* 90 degrees for starters, no obtuse angles */
  do{/* Search circle is defined by decreasing angle. */
    paramv->NumPoints=0;

    /* Radius of search circle */
    search_radius=HalfBase/sin(Angle); RadOffset=search_radius*cos(Angle);
    /* Center of search circle */
    centerX=MidX-(RadOffset *(HalfDeltaY/HalfBase));
    centerY=MidY+(RadOffset *(HalfDeltaX/HalfBase));
    SetCube(centerX-search_radius,centerY-search_radius,0.0,
            centerX+search_radius,centerY+search_radius,0.0,&szone);

    TraverseZone(tcontext->sort,tcontext->points,&szone,paramv);
    for (zcnt=0;zcnt<paramv->NumPoints;zcnt++){
      datindex=paramv->V[zcnt];
      {/* Triangular distance testing. */
        P=&(tcontext->points->P[datindex]);/* with-do for sample point. */

        /* Pythagorean Dist. */
        Dist=Hypot(P->Loc[0]-centerX,P->Loc[1]-centerY);
        if(Dist<=search_radius){/* must be inside search circle to compete */

          /* looking for maximum obtuseness in angle. */
          Spread=AngleSpread(
            BaseP0.Loc[0],BaseP0.Loc[1],/* p0x, p0y */
            P->Loc[0],P->Loc[1],/* testx, testy */
            BaseP1.Loc[0],BaseP1.Loc[1]);/* p1x, p1y */

          if(Spread<=Pi){/* if greater than 180 degrees, ignore. */
            if((pt0!=datindex)&&(pt1!=datindex)){/* don't find selves. */
              if(Spread>MaxAngle){
                if( (*Filter)(P) ){/* user-defined filter */
                  if (1){//in_salami(Tx0,Ty0,Tx1,Ty1,border_segment_max,P->Loc[0],P->Loc[1])){
                    min_point_index=datindex;
                    MaxAngle=Spread;
                    Found=true;
                  }
                }
              }
            }
          }
#if 0
          { double arm1, arm2;
            arm1=Hypot(BaseP0.Loc[0]-P->Loc[0],BaseP0.Loc[1]-P->Loc[1]);
            arm2=Hypot(BaseP1.Loc[0]-P->Loc[0],BaseP1.Loc[1]-P->Loc[1]);
            if (arm1>arm2)maxarm=arm1; else maxarm=arm2;
          }
#endif
        }
      }/* end of triangular distance testing. */

    }
    Angle/=2.0;/* Angle gets narrower with each expansion of search.*/
  }while (! (Found || (RadOffset>=border_segment_max) || (Angle<Pi/500.0)) );
  (*MinSubjIndex)=min_point_index;
  return(Found);
}/* End of FindThird. */


void PushLine(tin_contextT *tcon,lineT *newline)
/* *******************************************************************
*  Add line to line wall.
* ******************************************************************* */
{
  brickptrT Up,Novel,here;
  lineT *line;

  /* Add line to linetree. */
  linetree_insert(&(tcon->line_tree),newline);

  /* new brick */
  Novel=malloc(sizeof(brickT));
  Novel->linptr=newline; /* create and set line pointer. */

  here=tcon->loop_wall;

  /* Insert the new link behind me. */
  Up=here->prev; /* save link to my predecessor */
  Up->next=Novel; /* set prev's down pointer to novel */
  here->prev=Novel; /* set my up pointer to novel */

  Novel->prev=Up;/* link novel prev to up */
  Novel->next=here;/* link novel prev to up */

  tcon->loop_wall=here;
}/* PushLine */

void PopLine(tin_contextT *tcon)
/* *******************************************************************
* Take line out of line wall.
* ******************************************************************* */
{
  brickptrT up,down,here;
  lineT *line;
  here=tcon->loop_wall;
  line=here->linptr; up=here->prev; down=here->next;
  up->next=down; down->prev=up;

  free(here); here=down;
  linetree_remove(&(tcon->line_tree),line);
  tcon->loop_wall=here;
}/* PopLine */

#if 0
// depricated.
void DisposeAll(brickptrT *Place)
/* *******************************************************************
*  Return the whole line loop to heap memory, along with the lines
* it points to.
* ******************************************************************* */
{
  brickptrT now;
  now=(*Place);
  while(now->next!=now){
    PopLine(&now);
  }
  linetree_remove(&line_tree,now->linptr); free(now);
  (*Place)=NULL;
}/* DisposeAll */
#endif

void dispose_loop(brickptrT loop_wall)
/* *******************************************************************
*  Return the whole line loop to heap memory.
*  Disposes of the loop wall without touching the line tree.
* ******************************************************************* */
{
  brickptrT old,now;
  now=loop_wall;
  if (now!=NULL){
    do{
      old=now; now=now->next; free(old);
    }while(now!=loop_wall);
  }
}/* dispose_loop */

void kill_congruents(flexelT *line_tree,lineT *me)
/* *******************************************************************
* ******************************************************************* */
{
  linelistT *other_lines;
  lineT **L;
  uint32 pdex2,pdex1,nlines;
  int lcnt;

  pdex2=me->pdex2; pdex1=me->pdex1;
  other_lines=flexray_lookup(line_tree,pdex2);
  /* at this point, other_lines has returned a list of all
    the other lines that end where me-line begins. */
  if (other_lines!=NULL){
    nlines=other_lines->nlines; L=other_lines->L;
    for(lcnt=0;lcnt<nlines;lcnt++){
      /* if the pdex2 point is the same as my pdex1, we are congruent.
        So tag us both for extinction. */
      if (L[lcnt]->pdex2==pdex1){
        L[lcnt]->state=dead; me->state=dead;
      }
    }
  }
}/* kill_congruents */

bool MakeFirst(tin_contextT *tcontext)
/* *******************************************************************
*  Make first seed triangle.
* ******************************************************************* */
{
  intdex Cnt,vdex,num_points;
  intdex vert_index1=0, vert_index2=0, vert_index3=0;
  PListT *pl;
  XYZType *Point1,*Point2,*P,*VertPt;
  bool found=false;

  pl=tcontext->points; num_points=pl->NumPoints;
  P=pl->P;

  vert_index1=0; vert_index2=0; vert_index3=0;
  /* this is just to pick a random point with neighbors. */
  while (vert_index1<num_points){    
    found=false;

    Point1 = &(P[vert_index1]);
    
    if (Point1->ref!=vertexpnt){
      /*
        check if point1 used here.  If used, skip looking for the
        other two points and look for the next point1.
  
        If we go all the way from beginning to end and find nothing,
        declare the tinning to be over.
      */
      /* Find nearest one that's not self */
      found=FindClosest(tcontext,vert_index1,&vert_index2,&point_test);/* pythagorean closest. */
      if (found){
        Point2 = &(P[vert_index2]);
        /* Triangle search: Find point nearest to line. Always searches
        on the left hand side of line travel direction. */
        found=FindThird(tcontext,vert_index1,vert_index2,&vert_index3,&point_test);
      }
    }
    /* repeat until one triangle can be completed, or we run out of points. */
    if (found)break;

    vert_index1++;
  }/* end while */
  if (!found) return(false);/* No more tinning can be done. */

  VertPt = &(P[vert_index3]);

  /* triangle is counter-clockwise */
#if allegro
  draw_triangle(tcontext->points,vert_index1,vert_index2,vert_index3,4);
#endif
  store_triangle(tcontext,vert_index1,vert_index2,vert_index3);
  P[vert_index1].ref=vertexpnt;
  P[vert_index2].ref=vertexpnt;
  P[vert_index3].ref=vertexpnt;

  {/* with lines do make first triangle from lines. */
    lineT *newline;
    brickptrT loop_wall;

    /* Add first line to linetree. */
    newline=line_create(vert_index3,vert_index2,alive);
    if (tcontext->loop_wall==NULL){
      linetree_insert(&(tcontext->line_tree),newline);
      loop_wall=malloc(sizeof(brickT)); loop_wall->linptr = newline;
      loop_wall->prev=loop_wall; loop_wall->next=loop_wall;/* points to self. */
      tcontext->loop_wall=loop_wall;
    }else{
      PushLine(tcontext,newline);
    }
    /* Add second line to linetree. */
    newline=line_create(vert_index2,vert_index1,alive);
    PushLine(tcontext,newline);

    /* Add third line - complete first triangle */
    newline=line_create(vert_index1,vert_index3,alive);
    PushLine(tcontext,newline);

  }/* end with. */
  return(true);
}/* MakeFirst */

void tin_init(tin_contextT *tcontext,PListT *points)
/* *******************************************************************
*  Initialize, allocate, and sort TIN needs.
* ******************************************************************* */
{
  VectT *sort=malloc(sizeof(VectT));/* create vector headers. */
  VectT *paramv=malloc(sizeof(VectT));

  flexray_init(&(tcontext->line_tree));
  flexray_init(&(tcontext->tri_tree));

  plist_blankrefs(points);
  InitSortVect(sort,points->NumPoints);
  MakeBlankSortVect(paramv,points->NumPoints);

  jasort(points,sort);
  tcontext->loop_wall=NULL;
  tcontext->points=points; tcontext->sort=sort; tcontext->paramv=paramv;
  tcontext->num_triangles=0;
  /* default triangle leg size is practically infinity. */
  tcontext->border_segment_max=1.0/fudge;
//  tcontext->num_lines=0;
}/* tin_init */

void tin_destroy(tin_contextT *tcontext)
/* *******************************************************************
*  Free all of the TIN structures from memory.
* ******************************************************************* */
{
  FreeSortVect(tcontext->sort); FreeSortVect(tcontext->paramv);
  tcontext->points=NULL; tcontext->sort=NULL; tcontext->paramv=NULL;
  dispose_loop(tcontext->loop_wall); tcontext->loop_wall=NULL;
  flexray_destroy(&(tcontext->line_tree),&leaf_destroy);
  flexray_destroy(&(tcontext->tri_tree),&tri_destroy);
  tcontext->num_triangles=0;
  memset(tcontext,0,sizeof(tin_contextT));
//  tcontext->num_lines=0;
}/* tin_destroy */

void tin_create(tin_contextT *tcontext,PListT *points,double border_segment_max)
/* *******************************************************************
* ******************************************************************* */
{
  intdex pdex1,pdex2,VertDex,wallsize;
  bool Done=false,found=false;
  lineT *ThisLine,*newline,*outermark;
  XYZType *Point1,*Point2;
  VectT *sort;
  VectT paramv;

  tin_init(tcontext,points);
  /* set triangle leg size per the user/programmer`s request. */
  tcontext->border_segment_max=border_segment_max;

  sort=tcontext->sort;
  while (MakeFirst(tcontext)){
    Done=false;
    do{
      ThisLine=tcontext->loop_wall->linptr;
      kill_congruents(&(tcontext->line_tree),ThisLine);
  
      if(ThisLine->state==dead){
        /* if a wall line is dead, don't grow on it. */
        outermark=NULL;
        if(tcontext->loop_wall->next==tcontext->loop_wall){
          Done=true;/* if wall is down to 1 brick, we're done. */
        }else{
          PopLine(tcontext);/* Remove dead line from loop */
        }
      }else if (ThisLine->state==outer){
  
        /* Start checking if our *entire loop wall* is facing outer space. */
        if (outermark==NULL){
          outermark=ThisLine;
        }
  
        tcontext->loop_wall=tcontext->loop_wall->next;
        /* If we've gone completely around the loop and found only outers,
         then our job is done. */
        if(tcontext->loop_wall->linptr==outermark){
          Done=true;
        }
      }else{
        /* Else this wall segment is alive, search for room to expand. */
        VertDex=0;
        outermark=NULL;

        /* get base point to find third */
        pdex1=ThisLine->pdex1;  pdex2=ThisLine->pdex2;
        Point1=&(points->P[pdex2]); Point2=&(points->P[pdex1]);
  
        /* find nearest to line */
        /* always searches on the left hand side of line travel direction */
        found=FindThird(tcontext,pdex1,pdex2,&VertDex,&point_test);
        if(!found){
          /* Third not found, therefore this line is facing outer space. */
          ThisLine->state=outer;
          tcontext->loop_wall=tcontext->loop_wall->next;
        }else{
          /* Else we have a complete new triangle,
            save it and expand the loop wall. */

          /* triangle is counter-clockwise */
//          store_triangle(tcontext,pdex1,pdex2,VertDex);
          store_triangle(tcontext,pdex2,VertDex,pdex1);
          points->P[VertDex].ref=vertexpnt;
  
          /* Start modifying the loop wall. */
          ThisLine->state=dead;/* Remove base line from loop */
          PopLine(tcontext);

          /* Add two new lines where the old one was. */
          newline=line_create(pdex1,VertDex,alive);
          PushLine(tcontext,newline);

          newline=line_create(VertDex,pdex2,alive);
          PushLine(tcontext,newline);
        } /* end of ThisLine on outside - else */
      }/* end if ThisLine->state==dead - else */
    }while(!Done);
  }
}/* tin_create */


bool tin_read_xyzs(PListT *points,char *filename)
/* *******************************************************************
*  Read all the xyzs from the demo xyz.txt file into our point array.
*  Returns true if successful.
* ******************************************************************* */
{
  uint32 cnt,num_points;
  char bufstr[300];
  FILE *survey;
  XYZType *P;/* pointer for quick dereferencing. */
  bool retval=true;

  memset(bufstr,0,300);
  survey=fopen(filename,"r");
  if (survey==NULL){
    retval=false;/* Error opening file. */
  }else{
    /* first line is number of data xyz points. */
    fgets(bufstr,290,survey);
    /* validate that the first line matches our xyz format. */
    if (strstr(bufstr,"num_points")!=NULL){
      sscanf(bufstr," num_points=%li\n",&num_points);

      /* allocate the xyz array to fit the data size. */
      InitPList(points,num_points);

      cnt=0;
      while(cnt<num_points && fgets(bufstr,290,survey)){
        P=&(points->P[cnt]);
        sscanf(bufstr," %lf %lf %lf\n",&(P->Loc[0]),&(P->Loc[1]),&(P->Loc[2]));
        cnt++;
      }
      points->NumPoints=cnt;
      retval=true;
    }else{/* Doesn't match our xyz file format, lacks proper header. */
      retval=false;
    }
    fclose(survey);
  }
  return(retval);
}/* tin_read_xyzs */

