/* **************************************************************************

  TIN.H, a library to create Triangulated Irregular Networks
from xyz point data.

Copyright (C) 2000 John Conner Atkeson III

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

For further information, contact:
John Conner Atkeson III
jcatkeson at gmail dot com
http://jca3.freeshell.org
3959 Persimmon Drive #103 Fairfax Va. 22031

************************************************************************** */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <time.h>

#ifdef __DJGPP
  #include <dos.h>
  #define allegro 1
  #if allegro
    #include "allegro.h"
  #endif
#else
  #define allegro 0
#endif

#include "flexray.h"
#include "space.h"

#ifndef int8
  #define int8 char
  #define uint8 unsigned int8
#endif

#ifndef int32
  #define int32 int
  #define uint32 unsigned int32
#endif


#define byradius 1

// static double tin_radius=1.0/fudge;

/* possible states a line segment can have while we create a TIN. */
typedef enum{alive,dead,outer} lstateT;

typedef struct lineT{
  intdex pdex2,pdex1;/* indexes to endpoints */
  lstateT state;
}lineT;
/* an array of pointers to lines, nothing more. */
typedef struct linelistT{
  int nlines;
  lineT **L;
}linelistT;

/* triangle type. */
typedef struct triT{
  intdex vtex[3];/* indexes of 3 vertex points, in points array. */
}triT;

/* complete triangle type, no dereferencing required. */
typedef struct full_triT{
  double vtex[3][3];/* 3 xyz points */
  double slope[2];/* x and y slopes */
}full_triT;

/* Bricks in the loop wall. */
typedef struct brickT *brickptrT;
typedef struct brickT{
  brickptrT prev,next;
  lineT *linptr;
}brickT;

/* global vars */
// uint8 MaxDimension;
/*
PListT points;
VectT sort;
*/
typedef struct tin_contextT
/* *******************************************************************
*  Serves the purpose of global variables, but neater.
* ******************************************************************* */
{
  PListT *points;
  VectT *sort;
  VectT *paramv;

  /* top of flexray tree. */
  flexelT line_tree;
  /* running list of lines in boundary of growing lattice */
  brickptrT loop_wall;

  /* tree of triangles we create. */
  uint32 num_triangles;
  flexelT tri_tree;
  
  /* _roughly_ how long a triangle`s legs are allowed to be. */
  double border_segment_max;
}tin_contextT;

void tin_create(tin_contextT *tcontext,PListT *points,double border_segment_max);
void tin_destroy(tin_contextT *tcontext);
bool tin_get_triangle_full(tin_contextT *tcontext,full_triT *tri,uint32 tri_num);
void draw_triangle(PListT *points,intdex pdex1,intdex pdex2,intdex pdex3,int color);
void tin_draw_triangles(tin_contextT *tcontext);
bool tin_read_xyzs(PListT *points,char *filename);
void tin_save_triangles(tin_contextT *tcontext,char *filename);
void save_tinfile(tin_contextT *tcontext,char *filename);
void tin_save_outline(tin_contextT *tcontext,char *filename);
void DrawLoop(brickptrT loop_wall,PListT *points,int color);
uint32 tin_create_random(PListT *points,int numpoints);/* Generate random points for testing. */
